#!/bin/bash
# $Id: torque_setup.sh 529 2007-12-20 21:04:22Z lukasz $

main() {
	prefix=/opt/torque
	torque_user=torque
	if [ -z "$host" ]; then
		host=`hostname`
	fi
	if echo "$0" | grep -q "^/"; then
		script="$0"
	else
		script="`pwd`/$0"
	fi
	targets=()
	for i in "$@"; do
		case "$i" in
			--help|help)
				"usage: $0 [src=torque-tarball] [build|install]"
				;;
			*=*)
				eval "$i"
				;;
			*)
				targets=("${targets[@]}" "$i")
				;;
		esac
	done
	if [ ${#targets[@]} -eq 0 ]; then
		targets=(all)
	fi
	spool=${prefix}/spool
	export prefix spool torque_user host

	if make_ ${targets[@]}; then
		echo "success"
	else
		rc=$?
		echo "error; exiting with ${rc} code"
		return ${rc}
	fi
}


make_() {
	for target in "$@"; do
		echo "making ${target}..."
		case $target in
			all) make_ unpack build install ;;
			unpack) unpack ;;
			configure) configure ;;
			build) build ;;
			install) install ;;
			*)
				echo "unknown target ${target}" >&2
				return 1;
			;;
		esac
		rc=$?
		if [ \! $rc -eq 0 ]; then
			return ${rc}
		fi
	done
}


unpack() {
	echo "-> unpack"
	if [ -e ./configure ] \
			&& ./configure --help | head -1 | grep -q "configures torque"; then
		return 0
	fi
	if [ -z "$src" ]; then
		echo "source tarball not given append src=..." >&2
		return 1
	fi
	dir=`basename $src .tar.gz`
	if [ \! -d $dir ]; then
		echo "unpacking ${src}..."
		tar xfz "$src" || return 1
	fi
	cd $dir
	echo "<- unpack"
}


configure() {
	echo "-> configure"
	if [ \! -e configure.done ]; then
		echo "configuring Torque..."
		./configure --with-scp --prefix=${prefix} --with-server-home=${spool} \
		&& touch configure.done
	fi
	echo "<- configure"
}


build() {
	configure && make
}


install() {
	echo "-> install"
	if [ \! $EUID -eq 0 ]; then
		echo "gaining root priviledges: sudo ${script} install..."
		sudo ${script} install
		rc=$?
		echo "== after install"
		return $rc
	fi

	daemons="pbs_server pbs_mom pbs_sched"
	killall ${daemons} && sleep 2 \
		&& killall -9 ${daemons}

	rm -rf "${prefix}"
	make install || return 1

	PATH="${prefix}/bin:${prefix}/sbin:${PATH}"
	export PATH

	./torque.setup ${torque_user} || return 1

	killall pbs_server && sleep 2 && killall -9 pbs_server

	cat >${spool}/mom_priv/config <<EOF
\$pbs_server ${host}
EOF

	cat >${spool}/server_priv/nodes <<EOF
${host} np=4
EOF

	pbs_server && sleep 1 && pbs_mom && sleep 1 && pbs_sched

	qmgr <<EOF
set queue batch keep_completed = 120
EOF
	if [ -n "$USERNAME" -a "$USERNAME" != root ]; then
		user=$USERNAME
	elif [ -n "$SUDO_USER" -a "$SUDO_USER" != root ]; then
		user=$SUDO_USER
	fi
	if [ -n "$user" ]; then
		qmgr <<EOF
set server managers += ${user}@${host}
set server operators += ${user}@${host}
EOF
	fi
	echo "<- install"
}


main $@
